---
layout: default
---
<div class="subnav">
    <ul class="inline">
    {% for link in site.navigation %}
        {% if link.url == "/patterns/" %}
            {% for sublink in link.subs %}
                <li><a href="{{ sublink.url }}">{{ sublink.text }}</a></li>
            {% endfor %}
        {% endif %}
    {% endfor %}
    </ul>
</div>

{{ content }}
